package no.shhsoft.warchecker;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarInputStream;

import no.shhsoft.utils.OptionParser;
import no.shhsoft.utils.OptionParser.Opt;
import no.shhsoft.warchecker.reporters.AbstractReporter;
import no.shhsoft.warchecker.reporters.ContainerResourceOverrideReporter;
import no.shhsoft.warchecker.reporters.JarsReporter;
import no.shhsoft.warchecker.reporters.ResourceDuplicateReporter;
import no.shhsoft.warchecker.reporters.SimilarlyNamedJarsReporter;
import no.shhsoft.warchecker.reporters.SuspiciousJarsReporter;

/**
 * @author <a href="mailto:shh@thathost.com">Sverre H. Huseby</a>
 */
public final class WarChecker {

    private static final String LIB_PREFIX = "WEB-INF/lib/";
    private static final String CLASSES_PREFIX = "WEB-INF/classes/";
    private boolean includeJarsReporter = false;
    private final Out out;

    private static void addResourceFromEntry(final WarData warData, final JarEntry entry,
                                             final String nameToUse,
                                             final ResourceContainer container) {
        warData.addResource(nameToUse != null ? nameToUse : entry.getName(), entry.getSize(),
                            entry.getCrc(), entry.getTime(), container);
    }

    private static void processJar(final WarData warData, final InputStream inputStream,
                                   final String name, final boolean isExternal) {
        final ResourceContainer container = new ResourceContainer(name, isExternal);
        try {
            final JarInputStream in = new JarInputStream(inputStream);
            JarEntry entry;
            while ((entry = in.getNextJarEntry()) != null) {
                in.closeEntry();
                if (entry.isDirectory()) {
                    continue;
                }
                addResourceFromEntry(warData, entry, null, container);
            }
            in.close();
        } catch (final IOException e) {
            throw new RuntimeException("jar file read error: " + e.getMessage(), e);
        }
    }

    private static void processEmbeddedJar(final WarData warData, final InputStream inputStream,
                                           final String name) {
        processJar(warData, inputStream, name, false);
    }

    private void addExternalJar(final WarData warData, final String filename,
                                final boolean ignoreMissingFile) {
        try {
            final InputStream in = new BufferedInputStream(new FileInputStream(filename));
            processJar(warData, in, filename, true);
            in.close();
        } catch (final FileNotFoundException e) {
            if (ignoreMissingFile) {
                out.info("ignoring missing jar file `" + filename + "'");
            } else {
                throw new RuntimeException("missing jar file `" + filename + "'");
            }
        } catch (final IOException e) {
            throw new RuntimeException("error processing file", e);
        }
    }

    public static WarData loadWarFile(final String filename) {
        final WarData warData = new WarData(filename);
        try {
            final JarFile jf = new JarFile(filename);
            for (final Enumeration<JarEntry> e = jf.entries(); e.hasMoreElements();) {
                final JarEntry entry = e.nextElement();
                if (entry.isDirectory()) {
                    continue;
                }
                final String name = entry.getName();
                if (name.startsWith(LIB_PREFIX)) {
                    final String localName = name.substring(LIB_PREFIX.length());
                    processEmbeddedJar(warData, jf.getInputStream(entry), localName);
                } else if (name.startsWith(CLASSES_PREFIX)) {
                    final String nameToUse = name.substring(CLASSES_PREFIX.length());
                    addResourceFromEntry(warData, entry, nameToUse, null);
                }
            }
            return warData;
        } catch (final IOException e) {
            throw new RuntimeException("war file read error: " + e.getMessage(), e);
        }
    }

    private String format(final String s, final int indentLevel) {
        if (s == null || s.length() == 0) {
            return "";
        }
        final String newLine = System.getProperty("line.separator", "\n");
        final String indent = Indenter.getIndentSpaces(indentLevel);
        final StringBuilder sb = new StringBuilder();
        if (s.charAt(0) != '\n') {
            sb.append(indent);
        }
        for (int q = 0; q < s.length(); q++) {
            final char c = s.charAt(q);
            if (c == '\n') {
                sb.append(newLine);
                if (q + 1 < s.length() && s.charAt(q + 1) != '\n') {
                    sb.append(indent);
                }
            } else {
                sb.append(c);
            }
        }
        return sb.toString();
    }

    public WarChecker(final Out out) {
        this.out = out;
    }

    public void showUsage() {
        out.info(
            "\n"
            + "usage: WarChecker [options] file.war\n"
            + "\n"
            + "  -t, --compare-time    include time in resource detail comparison\n"
            + "  -n, --name-only       consider only the name when comparing resources\n"
            + "  -l, --list-jars       list all jar files\n"
        );
        System.exit(0);
    }

    public void enableSkipResourceComparison() {
        AbstractReporter.setSkipResourceComparison(true);
    }

    public void enableIncludeTimeInResourceComparison() {
        AbstractReporter.setIncludeTimeInResourceComparison(true);
    }

    public void enableJarsReporter() {
        includeJarsReporter = true;
    }

    public void checkWarFile(final String filename,
                             final List<String> additionalJars, final boolean ignoreMissingJar) {
        try {
            final WarData warData = loadWarFile(filename);
            if (additionalJars != null) {
                for (final String jarname : additionalJars) {
                    addExternalJar(warData, jarname, ignoreMissingJar);
                }
            }
            out.info("");
            if (includeJarsReporter) {
                out.info(format(new JarsReporter().report(warData), 1));
            }
            out.info(format(new SimilarlyNamedJarsReporter().report(warData), 1));
            out.info(format(new SuspiciousJarsReporter().report(warData), 1));
            out.info(format(new ContainerResourceOverrideReporter().report(warData), 1));
            out.info(format(new ResourceDuplicateReporter().report(warData), 1));
        } catch (final Throwable t) {
            out.error("error: " + t.getMessage());
        }
    }

    public void checkWarFile(final String filename) {
        checkWarFile(filename, null, true);
    }

    public static void main(final String[] args) {
        final WarChecker warChecker = new WarChecker(new ConsoleOut());
        final OptionParser parser = new OptionParser(new Opt[] {
            new Opt('h', "help", "showUsage"),
            new Opt('t', "compare-time", "enableIncludeTimeInResourceComparison"),
            new Opt('n', "name-only", "enableSkipResourceComparison"),
            new Opt('l', "list-jars", "enableJarsReporter"),
        });
        final String[] rest = parser.parse(args, warChecker);
        if (rest.length != 1) {
            warChecker.showUsage();
        }
        warChecker.checkWarFile(rest[0]);
    }

}
